#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import histbook.expr

import numpy
INDEXTYPE = numpy.int32

library = {}

library["numpy.add"] = numpy.add
library["numpy.subtract"] = numpy.subtract
library["numpy.multiply"] = numpy.multiply
library["numpy.true_divide"] = numpy.true_divide
library["numpy.equal"] = numpy.equal
library["numpy.not_equal"] = numpy.not_equal
library["numpy.less"] = numpy.less
library["numpy.less_equal"] = numpy.less_equal
try:
    library["numpy.isin"] = numpy.isin
except AttributeError:
    library["numpy.isin"] = numpy.in1d
library["numpy.logical_or"] = numpy.logical_or
library["numpy.logical_and"] = numpy.logical_and
library["numpy.logical_not"] = numpy.logical_not

library["abs"] = numpy.absolute;          library["fabs"] = numpy.absolute
library["arccos"] = numpy.arccos
library["arccosh"] = numpy.arccosh
library["arcsin"] = numpy.arcsin
library["arcsinh"] = numpy.arcsinh
library["arctan2"] = numpy.arctan2
library["arctan"] = numpy.arctan
library["arctanh"] = numpy.arctanh
library["ceil"] = numpy.ceil
library["conj"] = numpy.conjugate
library["copysign"] = numpy.copysign
library["cos"] = numpy.cos
library["cosh"] = numpy.cosh
library["deg2rad"] = numpy.deg2rad
library["exp2"] = numpy.exp2
library["exp"] = numpy.exp
library["expm1"] = numpy.expm1
library["floor"] = numpy.floor
library["fmod"] = numpy.fmod
try:
    library["heaviside"] = lambda x, middle=0.5: numpy.heaviside(x, middle)
except AttributeError:
    def heaviside(x, middle=0.5):
        out = numpy.where(x < 0, 0.0, 1.0)
        out[x == 0] = middle
        return out
    library["heaviside"] = heaviside
library["hypot"] = numpy.hypot
library["isfinite"] = numpy.isfinite
library["isinf"] = numpy.isinf
library["isnan"] = numpy.isnan
library["log10"] = numpy.log10
library["log1p"] = numpy.log1p
library["log2"] = numpy.log2
library["logaddexp2"] = numpy.logaddexp2
library["logaddexp"] = numpy.logaddexp
library["log"] = numpy.log
library["max"] = numpy.maximum;           library["fmax"] = numpy.maximum
library["min"] = numpy.minimum;           library["fmin"] = numpy.minimum
library["pow"] = numpy.power
library["rad2deg"] = numpy.rad2deg
library["mod"] = numpy.remainder;         library["fmod"] = numpy.remainder
library["rint"] = numpy.rint
library["sign"] = numpy.sign
library["sinh"] = numpy.sinh
library["sin"] = numpy.sin
library["sqrt"] = numpy.sqrt
library["tanh"] = numpy.tanh
library["tan"] = numpy.tan
library["trunc"] = numpy.trunc

def vectorized_erf(complement):
    a1 =  0.254829592
    a2 = -0.284496736
    a3 =  1.421413741
    a4 = -1.453152027
    a5 =  1.061405429
    p  =  0.3275911
    def erf(values):
        sign = numpy.where(values < 0, -1.0, 1.0)
        values = numpy.absolute(values)
        t = 1.0 / (values * p + 1)
        y = 1.0 - ((((a5*t + a4)*t + a3)*t + a2)*t + a1)*t * numpy.exp(numpy.negative(numpy.square(values)))
        if complement:
            return 1.0 - sign * y
        else:
            return sign * y
    return erf

library["erf"] = vectorized_erf(False)
library["erfc"] = vectorized_erf(True)

def vectorized_gamma(logarithm):
    cofs = (76.18009173, -86.50532033, 24.01409822, -1.231739516e0, 0.120858003e-2, -0.536382e-5)
    stp = 2.50662827465
    def lgamma(values):
        x = values - 1.0
        tmp = x + 5.5
        tmp = (x + 0.5)*numpy.log(tmp) - tmp
        ser = numpy.ones(len(values), dtype=numpy.dtype(numpy.float64))
        for cof in cofs:
            numpy.add(x, 1.0, x)
            numpy.add(ser, cof/x, ser)
        return tmp + numpy.log(stp*ser)
    if logarithm:
        return lgamma
    else:
        return lambda values: numpy.exp(lgamma(values))

library["gamma"] = vectorized_gamma(False)
lgamma = library["lgamma"] = vectorized_gamma(True)
library["factorial"] = lambda values: numpy.round(numpy.exp(lgamma(numpy.round(values) + 1)))

library["where"] = lambda condition, yes, no: numpy.where(condition, yes, no)

def histbook_groupby(values):
    uniques, inverse = numpy.unique(values, return_inverse=True)
    inverse = inverse.astype(INDEXTYPE)
    return uniques, inverse

library["histbook.groupby"] = histbook_groupby

def histbook_groupbin(nanflow, closedlow):
    def groupbin(values, binwidth, origin):
        if origin == 0:
            indexes = numpy.multiply(values, 1.0/float(binwidth))
        else:
            indexes = values - float(origin)
            numpy.multiply(indexes, 1.0/float(binwidth), indexes)

        if closedlow:
            numpy.floor(indexes, indexes)
        else:
            numpy.ceil(indexes, indexes)
            numpy.subtract(indexes, 1, indexes)

        numpy.multiply(indexes, float(binwidth), indexes)
        if origin != 0:
            numpy.add(indexes, float(origin), indexes)

        ok = numpy.isnan(indexes)
        numpy.logical_not(ok, ok)

        if ok.all():
            uniques, inverse = numpy.unique(indexes, return_inverse=True)
            inverse = inverse.astype(INDEXTYPE)

        else:
            uniques, okinverse = numpy.unique(indexes[ok], return_inverse=True)
            inverse = numpy.ones(indexes.shape, dtype=INDEXTYPE)

            if nanflow:
                numpy.multiply(inverse, len(uniques), inverse)
                inverse[ok] = okinverse
                uniques = list(uniques) + ["NaN"]

            else:
                numpy.multiply(inverse, -1, inverse)
                inverse[ok] = okinverse

        return uniques, inverse

    return groupbin

library["histbook.groupbinNL"] = histbook_groupbin(True, True)
library["histbook.groupbinNH"] = histbook_groupbin(True, False)
library["histbook.groupbin_L"] = histbook_groupbin(False, True)
library["histbook.groupbin_H"] = histbook_groupbin(False, False)
    
def histbook_bin(underflow, overflow, nanflow, closedlow):
    if nanflow:
        nanindex = (1 if underflow else 0) + (1 if overflow else 0)
    else:
        nanindex = numpy.ma.masked

    if underflow:
        shift = 1
    else:
        shift = 0

    def bin(values, numbins, low, high):
        indexes = values - float(low)
        numpy.multiply(indexes, float(numbins) / float(high - low), indexes)

        if closedlow:
            numpy.floor(indexes, indexes)
            if shift != 0:
                numpy.add(indexes, shift, indexes)
        else:
            numpy.ceil(indexes, indexes)
            numpy.add(indexes, shift - 1, indexes)

        out = numpy.ma.array(indexes, dtype=INDEXTYPE)
        with numpy.errstate(invalid="ignore"):
            if underflow:
                numpy.maximum(out, 0, out)
            else:
                out[out < 0] = numpy.ma.masked
            if overflow:
                numpy.minimum(out, shift + numbins, out)
            else:
                out[out >= (numbins + shift)] = numpy.ma.masked
            out[numpy.isnan(indexes)] = nanindex + numbins
        return out

    return bin

library["histbook.binUONL"] = histbook_bin(True, True, True, True)
library["histbook.binUONH"] = histbook_bin(True, True, True, False)
library["histbook.binUO_L"] = histbook_bin(True, True, False, True)
library["histbook.binUO_H"] = histbook_bin(True, True, False, False)
library["histbook.binU_NL"] = histbook_bin(True, False, True, True)
library["histbook.binU_NH"] = histbook_bin(True, False, True, False)
library["histbook.binU__L"] = histbook_bin(True, False, False, True)
library["histbook.binU__H"] = histbook_bin(True, False, False, False)
library["histbook.bin_ONL"] = histbook_bin(False, True, True, True)
library["histbook.bin_ONH"] = histbook_bin(False, True, True, False)
library["histbook.bin_O_L"] = histbook_bin(False, True, False, True)
library["histbook.bin_O_H"] = histbook_bin(False, True, False, False)
library["histbook.bin__NL"] = histbook_bin(False, False, True, True)
library["histbook.bin__NH"] = histbook_bin(False, False, True, False)
library["histbook.bin___L"] = histbook_bin(False, False, False, True)
library["histbook.bin___H"] = histbook_bin(False, False, False, False)

def histbook_intbin(underflow, overflow):
    if underflow:
        shift = 1
    else:
        shift = 0

    def intbin(values, min, max):
        indexes = numpy.ma.array((values + (shift - min)), dtype=INDEXTYPE)

        if underflow:
            numpy.maximum(indexes, 0, indexes)
        else:
            indexes[indexes < 0] = numpy.ma.masked

        if overflow:
            numpy.minimum(indexes, (shift + 1 + max - min), indexes)
        else:
            indexes[indexes > (shift + max - min)] = numpy.ma.masked

        return indexes

    return intbin

library["histbook.intbinUO"] = histbook_intbin(True, True)
library["histbook.intbinU_"] = histbook_intbin(True, False)
library["histbook.intbin_O"] = histbook_intbin(False, True)
library["histbook.intbin__"] = histbook_intbin(False, False)

def histbook_split(underflow, overflow, nanflow, closedlow):
    def split(values, edges):
        indexes = numpy.ma.array(numpy.digitize(values, edges), dtype=INDEXTYPE)
        if not closedlow:
            indexes[library["numpy.isin"](values, edges)] -= 1

        if not overflow:
            indexes[indexes == len(edges)] = numpy.ma.masked

        if nanflow:
            indexes[numpy.isnan(values)] = len(edges) + (1 if overflow else 0)
        else:
            indexes[numpy.isnan(values)] = numpy.ma.masked

        if not underflow:
            indexes[indexes == 0] = numpy.ma.masked
            numpy.subtract(indexes, 1, indexes)

        return indexes

    return split

library["histbook.splitUONL"] = histbook_split(True, True, True, True)
library["histbook.splitUONH"] = histbook_split(True, True, True, False)
library["histbook.splitUO_L"] = histbook_split(True, True, False, True)
library["histbook.splitUO_H"] = histbook_split(True, True, False, False)
library["histbook.splitU_NL"] = histbook_split(True, False, True, True)
library["histbook.splitU_NH"] = histbook_split(True, False, True, False)
library["histbook.splitU__L"] = histbook_split(True, False, False, True)
library["histbook.splitU__H"] = histbook_split(True, False, False, False)
library["histbook.split_ONL"] = histbook_split(False, True, True, True)
library["histbook.split_ONH"] = histbook_split(False, True, True, False)
library["histbook.split_O_L"] = histbook_split(False, True, False, True)
library["histbook.split_O_H"] = histbook_split(False, True, False, False)
library["histbook.split__NL"] = histbook_split(False, False, True, True)
library["histbook.split__NH"] = histbook_split(False, False, True, False)
library["histbook.split___L"] = histbook_split(False, False, False, True)
library["histbook.split___H"] = histbook_split(False, False, False, False)

library["histbook.cut"] = lambda values: numpy.ma.array(values, dtype=INDEXTYPE)

def calculate(expr, symbols):
    u"""
    Calculates an expression, given a dict of symbols to set values of named fields.

    Parameters
    ----------
    expr : :py:class:`Expr <histbook.expr.Expr>`
        expression to calculate

    symbols : dict of str \u2192 Numpy array
        values of :py:class:`Names <histbook.expr.Name>` and :py:class:`Predicates <histbook.expr.Predicate>` in the expression

    Returns
    -------
    Numpy array or number
    """

    if isinstance(expr, histbook.expr.BroadcastConst):
        return symbols[expr.name]

    elif isinstance(expr, (histbook.expr.Name, histbook.expr.Predicate)):
        return symbols[expr.value]

    elif isinstance(expr, histbook.expr.Const):
        return expr.value

    elif isinstance(expr, histbook.expr.Call) and expr.fcn in library:
        return library[expr.fcn](*(calculate(arg, symbols) for arg in expr.args))
            
    else:
        raise AssertionError(repr(expr))



# -*- coding: utf-8 -*-

import os.path

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = "1.0"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.graphviz",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["-templates"]

source_suffix = ".rst"
source_encoding = "utf-8"
master_doc = "index"

project = u"histbook"
copyright = u"2018, DIANA-HEP"
author = u"Jim Pivarski"

def get_version():
    g = {}
    exec(open(os.path.join("..", "..", "histbook", "version.py")).read(), g)
    return g["__version__"]

version = get_version()
release = get_version()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

add_function_parentheses = False

autoclass_content = "both"

# -- Options for HTML output ----------------------------------------------

# html_theme = "alabaster"
# html_theme_options = {}
# html_theme_path = []

html_title = u"histbook v" + get_version()

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["-static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not None, a "Last updated on:" timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to "%b %d, %Y".
html_last_updated_fmt = ""

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ""

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   "da", "de", "en", "es", "fi", "fr", "hu", "it", "ja"
#   "nl", "no", "pt", "ro", "ru", "sv", "tr", "zh"
# html_search_language = "en"

# A dictionary with options for the search language support, empty by default.
# "ja" uses this config value.
# "zh" user can custom change `jieba` dictionary path.
# html_search_options = {"type": "default"}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = "scorer.js"

# Output file base name for HTML help builder.
htmlhelp_basename = "histbookdoc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}
latex_documents = [(master_doc, "histbook.tex", u"histbook Documentation", u"Jim Pivarski", "manual"),]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# It false, will not define \strong, \code, 	itleref, \crossref ... but only
# \sphinxstrong, ..., \sphinxtitleref, ... To help avoid clash with user added
# packages.
# latex_keep_old_macro_names = True

# If false, no module index is generated.
# latex_domain_indices = True

# -- Options for manual page output ---------------------------------------

man_pages = [(master_doc, "histbook", u"histbook Documentation", [author], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [(master_doc, "histbook", u"histbook Documentation", author, "histbook", "One line description of project.", "Miscellaneous"),]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: "footnote", "no", or "inline".
# texinfo_show_urls = "footnote"

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ast
import functools
import itertools
import math
import sys
import types
try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy

import histbook.util.astunparse

class ExpressionError(Exception): pass

class Expr(object):
    """A symbolic expression, constructed from a lambda or string using :py:meth:`Expr.parse <histbook.expr.Expr.parse>`."""

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, ", ".join(self._reprargs()))

    # assumes __eq__ and __lt__ have been defined

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __gt__(self, other):
        return not self.__lt__(other) and not self.__eq__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __cmp__(self, other):
        if self.__eq__(other):
            return 0
        elif self.__lt__(other):
            return -1
        else:
            return 1

    @staticmethod
    def parse(expression, defs=None):
        """
        Convert a lambda or string ``expression`` into a :py:class:`Expr <histbook.expr.Expr>`.

        Parameters
        ----------
        expression : lambda or string
            expression involving constants (literals and ``pi``, ``e``, ``inf``, and ``nan``/``NaN``), field variables, and functions in ``Expr.recognized``; if a lambda, the fields are given by function parameters, if a string, any unrecognized names will be identified as fields

        defs : ``None`` or dict
            if not ``None``, provides names to recognize in ``expression`` (to avoid repetitive code)
        """
        _defs = {}
        if defs is not None:
            for n, x in defs.items():
                if isinstance(x, Expr):
                    _defs[n] = x
                elif sys.version_info[0] < 3 and isinstance(x, (str, unicode)):
                    _defs[n] = Expr.parse(x)
                elif sys.version_info[0] >= 3 and isinstance(x, str):
                    _defs[n] = Expr.parse(x)
                else:
                    try:
                        _defs[n] = Const(pickle.loads(pickle.dumps(x)))
                    except:
                        raise ExpressionError("object can't be included in an expression as symbol {0} because it refers to an unserializable object".format(repr(n)))

        pyast = ast.parse(expression, mode="eval").body

        calculate = {"+": lambda x, y: x + y,
                     "-": lambda x, y: x - y,
                     "*": lambda x, y: x * y,
                     "/": lambda x, y: float(x) / float(y),
                     "//": lambda x, y: int(x // y),
                     "%": lambda x, y: x % y,
                     "**": lambda x, y: x ** y,
                     "|": lambda x, y: numpy.uint64(x) | numpy.uint64(y),
                     "&": lambda x, y: numpy.uint64(x) & numpy.uint64(y),
                     "^": lambda x, y: numpy.uint64(x) ^ numpy.uint64(y)}

        env = dict(globals())
        def resolve(node):
            if isinstance(node, ast.Attribute):
                return getattr(resolve(node.value), node.attr, None)
            elif isinstance(node, ast.Name):
                return env.get(node.id, None)
            else:
                raise ExpressionError("functions must be named, not constructed: {0}".format(histbook.util.astunparse.tostring(node).strip()))

        def recurse(node, relations=False):
            if isinstance(node, ast.Num):
                return Const(node.n)

            elif isinstance(node, ast.Str):
                return Const(node.s)

            elif isinstance(node, ast.Dict) and len(node.keys) == 0:
                return Const(set())

            elif isinstance(node, ast.Set):
                content = [recurse(x) for x in node.elts]
                if all(isinstance(x, Const) for x in content):
                    return Const(set(x.value for x in content))
                else:
                    raise ExpressionError("sets in expressions may not contain variable contents: {0}".format(histbook.util.astunparse.tostring(node).strip()))

            elif isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
                if node.id == "None":
                    return Const(None)
                elif node.id == "True":
                    return Const(True)
                elif node.id == "False":
                    return Const(False)
                elif node.id in _defs:
                    return _defs[node.id]
                else:
                    return Name(node.id)

            elif isinstance(node, getattr(ast, "NameConstant", tuple)):  # node will never be tuple
                if node.value is None:
                    return Const(None)
                elif node.value is True:
                    return Const(True)
                elif node.value is False:
                    return Const(False)
                else:
                    raise AssertionError(node)

            elif relations and isinstance(node, ast.Compare) and len(node.ops) == 1:
                if   isinstance(node.ops[0], ast.Eq):    cmp, swap = "==",     None
                elif isinstance(node.ops[0], ast.NotEq): cmp, swap = "!=",     None
                elif isinstance(node.ops[0], ast.Lt):    cmp, swap = "<",      False
                elif isinstance(node.ops[0], ast.LtE):   cmp, swap = "<=",     False
                elif isinstance(node.ops[0], ast.Gt):    cmp, swap = "<",      True
                elif isinstance(node.ops[0], ast.GtE):   cmp, swap = "<=",     True
                elif isinstance(node.ops[0], ast.In):    cmp, swap = "in",     False
                elif isinstance(node.ops[0], ast.NotIn): cmp, swap = "not in", False
                else:
                    raise ExpressionError("only comparision relations supported: '==', '!=', '<', '<=', '>', '>=', 'in', and 'not in': {0}".format(histbook.util.astunparse.tostring(node).strip()))
                
                left = recurse(node.left)
                right = recurse(node.comparators[0])
                if swap is True:
                    left, right = right, left
                elif swap is None:
                    left, right = sorted([left, right])

                if (cmp == "in" or cmp == "not in") and not (isinstance(right, Const) and isinstance(right.value, set)):
                    raise ExpressionError("comparisons 'in' and 'not in' can only be used with a set: {0}".format(histbook.util.astunparse.tostring(node).strip()))

                return Relation(cmp, left, right)

            elif relations and isinstance(node, ast.Compare) and len(node.ops) > 1:
                return recurse(ast.BoolOp(ast.And(), [ast.Compare(node.left if i == 0 else node.comparators[i - 1], [node.ops[i]], [node.comparators[i]]) for i in range(len(node.ops))]), relations=True)

            elif isinstance(node, ast.Compare):
                raise ExpressionError("comparison operators are only allowed at the top of an expression: {0}".format(histbook.util.astunparse.tostring(node).strip()))

            elif relations and isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
                content = recurse(node.operand, relations=True)
                if isinstance(content, Name):
                    return Predicate(content.value, positive=False)
                else:
                    return Logical.negate(content).simplify()

            elif relations and isinstance(node, ast.BoolOp) and isinstance(node.op, ast.And):
                return functools.reduce(LogicalAnd.combine, [Logical.normalform(recurse(x, relations=True)) for x in node.values]).simplify()

            elif relations and isinstance(node, ast.BoolOp) and isinstance(node.op, ast.Or):
                return functools.reduce(LogicalOr.combine, [Logical.normalform(recurse(x, relations=True)) for x in node.values]).simplify()

            elif isinstance(node, ast.BoolOp):
                raise ExpressionError("logical operators are only allowed at the top of an expression: {0}".format(histbook.util.astunparse.tostring(node).strip()))
                
            elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.USub):
                content = recurse(node.operand)
                if isinstance(content, Const):
                    return Const(-content.value)
                else:
                    return PlusMinus.negate(content).simplify()

            elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.UAdd):
                return recurse(node.operand)

            elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Invert):
                content = recurse(node.operand)
                if isinstance(content, Const):
                    return Const(~content.value)
                else:
                    return BitAnd.negate(content).simplify()

            elif isinstance(node, ast.UnaryOp):
                raise ExpressionError("only unary operators supported: 'not', '-', '+', and '~': {0}".format(histbook.util.astunparse.tostring(node).strip()))

            elif isinstance(node, ast.BinOp):
                if   isinstance(node.op, ast.Add):      fcn = "+"
                elif isinstance(node.op, ast.Sub):      fcn = "-"
                elif isinstance(node.op, ast.Mult):     fcn = "*"
                elif isinstance(node.op, ast.Div):      fcn = "/"
                elif isinstance(node.op, ast.FloorDiv): op, fcn = "//", "numpy.floor_divide"
                elif isinstance(node.op, ast.Mod):      op, fcn = "%",  "fmod"
                elif isinstance(node.op, ast.Pow):      op, fcn = "**", "pow"
                elif isinstance(node.op, ast.BitOr):    fcn = "|"
                elif isinstance(node.op, ast.BitAnd):   fcn = "&"
                elif isinstance(node.op, ast.BitXor):   op, fcn = "^",  "xor"
                else:
                    raise ExpressionError("only binary operators supported: '+', '-', '*', '/', '//', '%', '**', '|', '&', and '^': {0}".format(histbook.util.astunparse.tostring(node).strip()))

                left = recurse(node.left)
                right = recurse(node.right)

                if isinstance(left, Const) and isinstance(right, Const):
                    return Const(calculate[fcn](left.value, right.value))

                if fcn == "+":
                    return PlusMinus.combine(left, right).simplify()
                elif fcn == "-":
                    return PlusMinus.combine(left, PlusMinus.negate(right)).simplify()
                elif fcn == "*":
                    return PlusMinus.distribute(left, right).simplify()
                elif fcn == "/":
                    # PlusMinus is implemented as a RingAlgebra, but it's really a Field
                    right = PlusMinus.normalform(right)

                    if len(right.pos) == len(right.neg) == 0:
                        negation = TimesDiv(TimesDiv.negateval(right.const), (), ())
                    elif right.const == PlusMinus.identity and len(right.pos) == 1 and len(right.neg) == 0:
                        negation = TimesDiv.negate(right.pos[0])
                    elif right.const == PlusMinus.identity and len(right.pos) == 0 and len(right.neg) == 1:
                        negation = TimesDiv.negate(right.neg[0])
                        negation.const = PlusMinus.negateval(negation.const)
                    else:
                        negation = TimesDiv.negate(right)   # additive terms in the denominator

                    return PlusMinus.distribute(left, negation).simplify()
                
                elif fcn == "pow" and isinstance(right, Const) and right.value == round(right.value) and 1 <= right.value < 5:
                    n = int(right.value)
                    left = PlusMinus.normalform(left)
                    if left.const == PlusMinus.identity and len(left.pos) == 1 and len(left.neg) == 0:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(left.pos[0].const**n, tuple(sorted(left.pos[0].pos * n)), tuple(sorted(left.pos[0].neg * n))),), ()).simplify()
                    elif left.const == PlusMinus.identity and len(left.pos) == 0 and len(left.neg) == 1 and n % 2 == 0:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(left.neg[0].const**n, tuple(sorted(left.neg[0].pos * n)), tuple(sorted(left.neg[0].neg * n))),), ()).simplify()
                    elif left.const == PlusMinus.identity and len(left.pos) == 0 and len(left.neg) == 1:
                        return PlusMinus(PlusMinus.identity, (), (TimesDiv(left.neg[0].const**n, tuple(sorted(left.neg[0].pos * n)), tuple(sorted(left.neg[0].neg * n))),)).simplify()
                    else:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(TimesDiv.identity, tuple((left,) * n), ()),), ()).simplify()

                elif fcn == "pow" and isinstance(right, Const) and right.value == round(right.value) and -5 < right.value <= -1:
                    n = -int(right.value)
                    left = PlusMinus.normalform(left)
                    if left.const == PlusMinus.identity and len(left.pos) == 1 and len(left.neg) == 0:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(left.pos[0].const**(-n), tuple(sorted(left.pos[0].neg * n)), tuple(sorted(left.pos[0].pos * n))),), ()).simplify()
                    elif left.const == PlusMinus.identity and len(left.pos) == 0 and len(left.neg) == 1 and n % 2 == 0:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(left.neg[0].const**(-n), tuple(sorted(left.neg[0].neg * n)), tuple(sorted(left.neg[0].pos * n))),), ()).simplify()
                    elif left.const == PlusMinus.identity and len(left.pos) == 0 and len(left.neg) == 1:
                        return PlusMinus(PlusMinus.identity, (), (TimesDiv(left.neg[0].const**(-n), tuple(sorted(left.neg[0].neg * n)), tuple(sorted(left.neg[0].pos * n))),)).simplify()
                    else:
                        return PlusMinus(PlusMinus.identity, (TimesDiv(TimesDiv.identity, (), tuple((left,) * n)),), ()).simplify()

                elif fcn == "&":
                    raise NotImplementedError

                elif fcn == "|":
                    raise NotImplementedError

                else:
                    return BinOp(fcn, left, right, op)

            elif isinstance(node, ast.Call):
                fcn = Expr.recognized.get(resolve(node.func), None)

                if fcn is None and node.func.id in Expr.recognized.values():
                    fcn = node.func.id

                elif fcn is not None:
                    for n, x in Expr.recognized.items():
                        if fcn is n:
                            fcn = x
                            break
                    
                if fcn is None:
                    raise ExpressionError("unhandled function in expression: {0}".format(histbook.util.astunparse.tostring(node).strip()))

                return Call(fcn, *(recurse(x, relations=(i == 0 and fcn == "where")) for i, x in enumerate(node.args)))

            else:
                ExpressionError("unhandled syntax in expression: {0}".format(histbook.util.astunparse.tostring(node).strip()))

        return recurse(pyast, relations=True)
        
    maybeconstants = {"pi": numpy.pi, "Pi": numpy.pi,
                      "e": numpy.e, "E": numpy.e,
                      "inf": numpy.inf, "Inf": numpy.inf, "infinity": numpy.inf, "Infinity": numpy.inf,
                      "nan": numpy.nan, "NaN": numpy.nan, "Nan": numpy.nan}

    recognized = {abs: "abs", max: "max", min: "min"}

class _Placeholder(object):
    count = 0
    def __init__(self):
        self.index = _Placeholder.count
        _Placeholder.count += 1
    def __repr__(self):
        return "Placeholder({0})".format(self.index)
    def __hash__(self):
        return hash((_Placeholder, self.index))
    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.index == other.index
    def __ne__(self, other):
        return not self.__eq__(other)

def _recognize(module, pyname, name):
    if hasattr(module, pyname):
        Expr.recognized[getattr(module, pyname)] = name
    else:
        Expr.recognized[_Placeholder()] = name

Expr.recognized[_Placeholder()] = "fabs"
Expr.recognized[_Placeholder()] = "fmin"
Expr.recognized[_Placeholder()] = "fmax"
Expr.recognized[_Placeholder()] = "fmod"

_recognize(math, "acos", "arccos")
_recognize(math, "acosh", "arccosh")
_recognize(math, "asin", "arcsin")
_recognize(math, "asinh", "arcsinh")
_recognize(math, "atan2", "arctan2")
_recognize(math, "atan", "arctan")
_recognize(math, "atanh", "arctanh")
_recognize(math, "ceil", "ceil")
_recognize(math, "copysign", "copysign")
_recognize(math, "cos", "cos")
_recognize(math, "cosh", "cosh")
_recognize(math, "degrees", "rad2deg")    # not in numexpr
_recognize(math, "erfc", "erfc")
_recognize(math, "erf", "erf")
_recognize(math, "exp", "exp")
_recognize(math, "expm1", "expm1")
_recognize(math, "factorial", "factorial")    # not in numexpr
_recognize(math, "floor", "floor")
_recognize(math, "fmod", "mod")
_recognize(math, "gamma", "gamma")    # not in numexpr (has lgamma and tgamma)
_recognize(math, "hypot", "hypot")
_recognize(math, "isinf", "isinf")
_recognize(math, "isnan", "isnan")
_recognize(math, "lgamma", "lgamma")
_recognize(math, "log10", "log10")
_recognize(math, "log1p", "log1p")
_recognize(math, "log", "log")
_recognize(math, "pow", "pow")    # not in numexpr (only as operator **)
_recognize(math, "radians", "deg2rad")    # not in numexpr
_recognize(math, "sinh", "sinh")
_recognize(math, "sin", "sin")
_recognize(math, "sqrt", "sqrt")
_recognize(math, "tanh", "tanh")
_recognize(math, "tan", "tan")
_recognize(math, "trunc", "trunc")

_recognize(numpy, "absolute", "abs")
_recognize(numpy, "arccos", "arccos")
_recognize(numpy, "arccosh", "arccosh")
_recognize(numpy, "arcsin", "arcsin")
_recognize(numpy, "arcsinh", "arcsinh")
_recognize(numpy, "arctan2", "arctan2")
_recognize(numpy, "arctan", "arctan")
_recognize(numpy, "arctanh", "arctanh")
_recognize(numpy, "ceil", "ceil")
_recognize(numpy, "conjugate", "conj")
_recognize(numpy, "copysign", "copysign")
_recognize(numpy, "cos", "cos")
_recognize(numpy, "cosh", "cosh")
_recognize(numpy, "deg2rad", "deg2rad")    # not in numexpr
_recognize(numpy, "degrees", "rad2deg")    # not in numexpr
_recognize(numpy, "exp2", "exp2")
_recognize(numpy, "exp", "exp")
_recognize(numpy, "expm1", "expm1")
_recognize(numpy, "floor", "floor")
_recognize(numpy, "fmod", "fmod")
_recognize(numpy, "heaviside", "heaviside")    # not in numexpr
_recognize(numpy, "hypot", "hypot")
_recognize(numpy, "isfinite", "isfinite")
_recognize(numpy, "isinf", "isinf")
_recognize(numpy, "isnan", "isnan")
_recognize(numpy, "log10", "log10")
_recognize(numpy, "log1p", "log1p")
_recognize(numpy, "log2", "log2")
_recognize(numpy, "logaddexp2", "logaddexp2")    # not in numexpr
_recognize(numpy, "logaddexp", "logaddexp")    # not in numexpr
_recognize(numpy, "log", "log")
_recognize(numpy, "maximum", "fmax")
_recognize(numpy, "minimum", "fmin")
_recognize(numpy, "power", "pow")    # not in numexpr (only as operator **)
_recognize(numpy, "rad2deg", "rad2deg")    # not in numexpr
_recognize(numpy, "radians", "deg2rad")    # not in numexpr
_recognize(numpy, "remainder", "mod")
_recognize(numpy, "rint", "rint")
_recognize(numpy, "round", "round")
_recognize(numpy, "sign", "sign")    # not in numexpr
_recognize(numpy, "sinh", "sinh")
_recognize(numpy, "sin", "sin")
_recognize(numpy, "sqrt", "sqrt")
_recognize(numpy, "tanh", "tanh")
_recognize(numpy, "tan", "tan")
_recognize(numpy, "trunc", "trunc")
_recognize(numpy, "where", "where")

### in numexpr, but not defined here (TODO)
# "abs2"
# "bool"
# "cast"
# "cbrt"
# "complex"
# "copy"
# "crosspower"
# "fdim"
# "float32"
# "float64"
# "fma"
# "fpclassify"
# "ilogb"
# "imag"
# "int16"
# "int32"
# "int64"
# "int8"
# "isnormal"
# "lgamma"
# "logb"
# "logical_and"
# "logical_or"
# "lrint"
# "lround"
# "nearbyint"
# "neg"
# "nextafter"
# "nexttoward"
# "ones_like"
# "real"
# "scalbln"
# "signbit"
# "tgamma"
# "uint16"
# "uint32"
# "uint64"
# "uint8"
# "unsafe_cast"

class Const(Expr):
    """Represents a literal constant in the expression tree, such as a number, boolean, or ``None``."""

    def __init__(self, value):
        self.value = value

    def _reprargs(self):
        return (repr(self.value),)

    def __str__(self):
        return str(self.value)

    def __hash__(self):
        if isinstance(self.value, set):
            value = (set, tuple(sorted(self.value)))
        else:
            value = self.value
        return hash((Const, value))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.value == other.value

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            if type(self.value).__name__ == type(other.value).__name__:
                return self.value < other.value
            else:
                return type(self.value).__name__ < type(other.value).__name__
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def rename(self, names):
        return self

class Name(Expr):
    """Represents a named variable in the expression tree."""

    def __init__(self, value):
        self.value = value

    def _reprargs(self):
        return (repr(self.value),)

    def __str__(self):
        return self.value

    def __hash__(self):
        return hash((Name, self.value))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.value == other.value

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return self.value < other.value
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def rename(self, names):
        assert self in names
        return Name(names[self])

class BroadcastConst(Name):
    """Represents a literal constant in the expression tree that should be broadcast to the length of field data."""

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def _reprargs(self):
        return (repr(self.name), repr(self.value))

    def __str__(self):
        if self.name is None:
            return str(self.value)
        else:
            return self.name

    def __hash__(self):
        if isinstance(self.value, set):
            value = (set, tuple(sorted(self.value)))
        else:
            value = self.value
        return hash((Const, self.name, value))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.name == other.name and self.value == other.value

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            if type(self.name).__name__ == type(other.name).__name__:
                if self.name == other.name:
                    if type(self.value).__name__ == type(other.value).__name__:
                        return self.value < other.value
                    else:
                        return type(self.value).__name__ < type(other.value).__name__
                else:
                    self.name < other.name
            else:
                return type(self.name).__name__ < type(other.name).__name__
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def rename(self, names):
        assert self in names
        return BroadcastConst(names[self], self.value)

class Call(Expr):
    """Represents a function call in the expression tree."""

    def __init__(self, fcn, *args):
        self.fcn = fcn
        self.args = args

    def _reprargs(self):
        return (repr(self.fcn),) + tuple(repr(x) for x in self.args)

    def __str__(self):
        return "{0}({1})".format(self.fcn, ", ".join(str(x) for x in self.args))

    def __hash__(self):
        return hash((Call, self.fcn, self.args))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.fcn == other.fcn and self.args == other.args

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return (self.fcn, self.args) < (other.fcn, other.args)
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def rename(self, names):
        if self in names:
            return Name(names[self])
        else:
            return self.__class__(self.fcn, *(x.rename(names) for x in self.args))

class BinOp(Call):
    """Represents a binary operation in the expression tree, such as ``%`` or ``**``."""

    def __init__(self, fcn, left, right, op):
        super(BinOp, self).__init__(fcn, left, right)
        self.op = op

    def __str__(self):
        return (" " + self.op + " ").join(("(" + str(x) + ")") if isinstance(x, BinOp) else str(x) for x in self.args)

class RingAlgebra(Expr):
    """Abstract class for ring algebras, such as addition and multiplication."""

    def __init__(self, const, pos, neg):
        self.const = const
        self.pos = pos
        self.neg = neg

    def _reprargs(self):
        return (repr(self.const), repr(self.pos), repr(self.neg))

    def __hash__(self):
        return hash((self.__class__, self.const, self.pos, self.neg))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.const == other.const and self.pos == other.pos and self.neg == other.neg

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return (self.const, self.pos, self.neg) < (other.const, other.pos, other.neg)
        else:
            return self.__class__.__name__ < other.__class__.__name__

    commutative = False

    @classmethod
    def negate(op, arg):
        arg = op.normalform(arg)
        return op(op.negateval(arg.const), arg.neg, arg.pos)

    @classmethod
    def collect(op, arg):
        return arg

    @classmethod
    def combine(op, left, right):
        left, right = op.normalform(left), op.normalform(right)

        if op.commutative:
            pos = tuple(sorted(left.pos + right.pos))
            neg = tuple(sorted(left.neg + right.neg))
        else:
            pos = left.pos + right.pos
            neg = left.neg + right.neg

        return op.collect(op(op.calcval(left.const, right.const), pos, neg))

    def rename(self, names):
        if self in names:
            return Name(names[self])
        else:
            return self.__class__(self.const, tuple(x.rename(names) for x in self.pos), tuple(x.rename(names) for x in self.neg))

class RingAlgebraMultLike(RingAlgebra):
    """Abstract class for the multiplication-like part of a ring algebra."""

    @classmethod
    def normalform(op, arg):
        if isinstance(arg, op):
            return arg
        elif isinstance(arg, Const):
            return op(arg.value, (), ())
        else:
            return op(op.identity, (arg,), ())

    def similar(self, other):
        return isinstance(other, self.__class__) and self.pos == other.pos and self.neg == other.neg
        
class RingAlgebraAddLike(RingAlgebra):
    """Abstract class for the addition-like part of a ring algebra."""

    @classmethod
    def normalform(op, arg):
        if isinstance(arg, op):
            return arg
        else:
            arg = op.subop.normalform(arg)
            if len(arg.pos) == len(arg.neg) == 0:
                return op(arg.const, (), ())
            elif op.isnegval(arg.const):
                return op(op.identity, (), (op.subop(op.negateval(arg.const), arg.pos, arg.neg),))
            else:
                return op(op.identity, (arg,), ())

    @classmethod
    def collect(op, arg):
        const = arg.const
        terms = []

        def cancel(x):
            pos, neg = list(x.pos), []
            for y in x.neg:
                try:
                    i = x.pos.index(y)
                except ValueError:
                    neg.append(y)
                else:
                    del pos[i]
            return op.subop(x.const, tuple(pos), tuple(neg))

        for x in arg.pos:
            x = cancel(x)
            if x.const == op.identity:
                pass
            elif len(x.pos) == len(x.neg) == 0:
                const = op.calcval(const, x.const)
            else:
                for y in terms:
                    if x.similar(y):
                        y.const = op.calcval(x.const, y.const)
                        break
                else:
                    terms.append(x)

        for x in arg.neg:
            x = cancel(x)
            if x.const == op.identity:
                pass
            elif len(x.pos) == len(x.neg) == 0:
                const = op.calcval(const, op.negateval(x.const))
            else:
                for y in terms:
                    if x.similar(y):
                        y.const = op.calcval(op.negateval(x.const), y.const)
                        break
                else:
                    terms.append(op.subop(op.negateval(x.const), x.pos, x.neg))

        posterms, negterms = [], []
        for x in terms:
            if x.const == op.identity:
                pass
            elif op.isnegval(x.const):
                x.const = op.negateval(x.const)
                negterms.append(x)
            else:
                posterms.append(x)

        if op.commutative:
            posterms.sort()
            negterms.sort()

        return op(const, tuple(posterms), tuple(negterms))
        
    @classmethod
    def distribute(op, left, right):
        left, right = op.normalform(left), op.normalform(right)
        pos, neg = [], []

        pos += [op.subop.combine(x, Const(right.const)) for x in left.pos]
        neg += [op.subop.combine(x, Const(right.const)) for x in left.neg]
        pos += [op.subop.combine(Const(left.const), y) for y in right.pos]
        neg += [op.subop.combine(Const(left.const), y) for y in right.neg]

        pos += [op.subop.combine(x, y) for x, y in itertools.product(left.pos, right.pos)]
        neg += [op.subop.combine(x, y) for x, y in itertools.product(left.pos, right.neg)]
        pos += [op.subop.combine(x, y) for x, y in itertools.product(left.neg, right.neg)]
        neg += [op.subop.combine(x, y) for x, y in itertools.product(left.neg, right.pos)]

        return op.collect(op(op.identity, tuple(pos), tuple(neg)))

    def simplify(self):
        out = self.__class__(self.const, self.pos, self.neg)

        if len(out.pos) == len(out.neg) == 0:
            return Const(out.const)

        elif out.const == out.identity and len(out.pos) + len(out.neg) == 1:
            if len(out.pos) == 1:
                out = out.pos[0]
            else:
                out = self.subop(self.negateval(out.neg[0].const), out.neg[0].pos, out.neg[0].neg)

            if len(out.pos) == len(out.neg) == 0:
                return Const(out.const)
            elif out.const == out.identity and len(out.pos) == 1 and len(out.neg) == 0:
                return out.pos[0]

        return out

class RingAlgebraBinOp(object):
    """Abstract class for a binary operation in a ring algebra (for pretty printing)."""

    def __str__(self):
        out = []
        if self.const != self.identity or len(self.pos) == 0:
            out.append(repr(self.const))
        for x in self.pos:
            if len(out) > 0:
                out.append(self.posop)
            out.append(str(x))
        for x in self.neg:
            out.append(self.negop)
            out.append(str(x))
        return "".join(out)

class TimesDiv(RingAlgebraBinOp, RingAlgebraMultLike):
    """Represents multiplication and division in a ring algebra (actually, a whole field because it includes division)."""

    posop = "*"
    negop = "/"

    commutative = True
    identity = 1

    @staticmethod
    def negateval(value):
        return 1.0 / value            # only because arithmetic is a field, not a ring

    @staticmethod
    def calcval(left, right):
        return left * right

class PlusMinus(RingAlgebraBinOp, RingAlgebraAddLike):
    """Represents addition and subtraction in a ring algebra (actually, a whole field because it includes division)."""

    posop = " + "
    negop = " - "

    subop = TimesDiv
    commutative = True
    identity = 0

    @staticmethod
    def negateval(value):
        return -value

    @staticmethod
    def isnegval(value):
        return value < 0

    @staticmethod
    def calcval(left, right):
        return left + right

### FIXME: NotImplementedError

# class BitAnd(RingAlgebraMultLike):
#     commutative = True
#     identity = numpy.uint64(-1)

#     @staticmethod
#     def negateval(value):
#         raise AssertionError(value)   # bitwise arithmetic really is a ring

#     @staticmethod
#     def calcval(left, right):
#         return numpy.uint64(left) & numpy.uint64(right)

# class BitOr(RingAlgebraAddLike):
#     subop = BitAnd
#     commutative = True
#     identity = numpy.uint64(0)

#     @staticmethod
#     def negateval(value):
#         return ~numpy.uint64(value)

#     @staticmethod
#     def isnegval(value):
#         return False

#     @staticmethod
#     def calcval(left, right):
#         return numpy.uint64(left) | numpy.uint64(right)

class Logical(object):
    """Mix-in for logical operations."""

    commutative = True

    def __init__(self, *args):
        self.args = tuple(sorted(set(args)))

    def _reprargs(self):
        return tuple(repr(x) for x in self.args)

    def __hash__(self):
        return hash((self.__class__, self.args))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.args == other.args

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return self.args < other.args
        else:
            return self.__class__.__name__ < other.__class__.__name__

    @classmethod
    def normalform(cls, arg):
        if isinstance(arg, Name):
            return LogicalOr(LogicalAnd(Predicate(arg.value)))
        elif isinstance(arg, Predicate):
            return LogicalOr(LogicalAnd(arg))
        elif isinstance(arg, Relation):
            return LogicalOr(LogicalAnd(arg))
        elif isinstance(arg, LogicalAnd):
            return LogicalOr(arg)
        elif isinstance(arg, LogicalOr):
            return arg
        else:
            raise AssertionError(arg)

    @classmethod
    def negate(op, arg):
        logicalor = op.normalform(arg)
        upsidedown = [[relation.negate() for relation in logicaland.args] for logicaland in logicalor.args]
        return LogicalOr(*(LogicalAnd(*x) for x in itertools.product(*upsidedown)))

    def rename(self, names):
        if self in names:
            return Name(names[self])
        else:
            return self.__class__(*(x.rename(names) for x in self.args))

class LogicalAnd(Logical, RingAlgebraMultLike):
    """Represents logical and in a ring algebra."""

    @classmethod
    def combine(op, left, right):
        left, right = op.normalform(left), op.normalform(right)
        return LogicalOr(*(LogicalAnd(*(x.args + y.args)) for x, y in itertools.product(left.args, right.args)))

    def simplify(self):
        if len(self.args) == 0 or all(x == Const(True) for x in self.args):
            return Const(True)
        elif any(x == Const(False) for x in self.args):
            return Const(False)
        elif len(self.args) == 1:
            return self.args[0]
        else:
            return self

class LogicalOr(Logical, RingAlgebraAddLike):
    """Represents logical or in a ring algebra."""

    @classmethod
    def combine(op, left, right):
        left, right = op.normalform(left), op.normalform(right)
        return LogicalOr(*(left.args + right.args))

    def simplify(self):
        if len(self.args) == 0 or all(x == Const(False) for x in self.args):
            return Const(False)
        elif any(x == Const(True) for x in self.args):
            return Const(True)
        elif len(self.args) == 1:
            return self.args[0].simplify()
        else:
            return self

class Relation(Expr):
    """Represents a logical relation in the expression tree, such as ``==`` and ``in``."""

    def __init__(self, cmp, left, right):
        self.cmp = cmp
        self.left = left
        self.right = right

    def _reprargs(self):
        return (repr(self.cmp), repr(self.left), repr(self.right))

    def __str__(self):
        return "{0} {1} {2}".format(str(self.left), self.cmp, str(self.right))

    def __hash__(self):
        return hash((Relation, self.cmp, self.left, self.right))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.cmp == other.cmp and self.left == other.left and self.right == other.right

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return (self.cmp, self.left, self.right) < (other.cmp, other.left, other.right)
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def negate(self):
        if self.cmp == "==":
            return Relation("!=", self.left, self.right)
        elif self.cmp == "!=":
            return Relation("==", self.left, self.right)
        elif self.cmp == "<":
            return Relation("<=", self.right, self.left)
        elif self.cmp == "<=":
            return Relation("<", self.right, self.left)
        elif self.cmp == "in":
            return Relation("not in", self.left, self.right)
        elif self.cmp == "not in":
            return Relation("in", self.left, self.right)
        else:
            raise AssertionError(self.cmp)

class Predicate(Expr):
    """Represents a logical predicate (name identified as boolean) in the expression tree."""

    def __init__(self, value, positive=True):
        self.value = value
        self.positive = positive

    def _reprargs(self):
        return (repr(self.value), repr(self.positive))

    def __str__(self):
        if self.positive:
            return self.value
        else:
            return "not " + self.value

    def __hash__(self):
        return hash((Predicate, self.value, self.positive))

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self.value == other.value and self.positive == other.positive

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return (self.value, self.positive) < (other.value, other.positive)
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def negate(self):
        return Predicate(self.value, positive=not self.positive)

    def rename(self, names):
        assert self in names
        return Predicate(names[self])



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy

import histbook.calc
import histbook.expr
import histbook.instr
import histbook.util

class Fillable(object):
    """Mix-in for objects with a ``fill`` method, like `Hist <histbook.hist.Hist>` and `Book <histbook.hist.Book>`."""

    @property
    def fields(self):
        """Names of fields that must be provided in the ``fill`` method."""

        if self._fields is None:
            table = {}
            goals = set(self._goals)

            for x in goals:
                x.clear()
            for x in goals:
                x.grow(table)
            
            fields = histbook.instr.sources(goals, table)

            self._instructions = self._streamline(0, list(histbook.instr.instructions(fields, goals)))
            self._fields = sorted(x.goal.value for x in fields)

        return self._fields

    def _showgoals(self):
        self.fields  # for the side-effect of creating self._instructions

        numbers = {}
        order = []
        def recurse(node):
            for x in node.requires:
                recurse(x)
            if node not in numbers:
                number = numbers[node] = len(numbers)
                order.append(node)
        for goal in sorted(self._goals):
            recurse(goal)
        print("goals:")
        print("------")
        for node in order:
            print("#{0:<3d} requires {1:<10s} requiredby {2:<10s} ({3} total) for {4}".format(numbers[node], " ".join(map(repr, sorted(numbers[x] for x in node.requires))), " ".join(map(repr, sorted(numbers[x] for x in node.requiredby))), node.numrequiredby, repr(str(node.goal))))
        print("")
        print("instructions:")
        print("-------------")
        for instruction in self._instructions:
            print(instruction)
        print("")
        
    def _fill(self, arrays):
        self.fields  # for the side-effect of creating self._instructions
        
        length = None
        firstinstruction = None
        firstarray = None
        for instruction in self._instructions:
            if isinstance(instruction, histbook.instr.Param) and not isinstance(instruction.extern, histbook.expr.BroadcastConst):
                try:
                    array = arrays[instruction.extern.value]
                except KeyError:
                    if instruction.extern.value in histbook.expr.Expr.maybeconstants:
                        continue
                    else:
                        raise ValueError("required field {0} not found in fill arguments".format(repr(str(instruction.extern))))

                if not isinstance(array, numpy.ndarray):
                    array = numpy.array(array)
                if array.shape != ():
                    length = array.shape[0]
                    firstinstruction = instruction.name
                    firstarray = array
                    break

        if length is None:
            length = 1

        def full(length, value):
            if isinstance(value, (histbook.util.string, bytes)):
                value = numpy.array(value)
            if isinstance(value, numpy.ndarray) and value.shape == () and (value.dtype.kind == "U" or value.dtype.kind == "S"):
                out = numpy.empty(length, dtype=value.dtype)
                out[:] = value
                return out
            else:
                return numpy.full(length, value)

        symbols = {}
        for instruction in self._instructions:
            if isinstance(instruction, histbook.instr.Param):
                if isinstance(instruction.extern, histbook.expr.BroadcastConst):
                    array = full(length, instruction.extern.value)
                elif instruction.name == firstinstruction:
                    array = firstarray
                else:
                    try:
                        array = arrays[instruction.extern.value]
                    except KeyError:
                        if instruction.extern.value in histbook.expr.Expr.maybeconstants:
                            array = full(length, histbook.expr.Expr.maybeconstants[instruction.extern.value])
                        else:
                            raise ValueError("required field {0} not found in fill arguments".format(repr(str(instruction.extern))))

                if not isinstance(array, numpy.ndarray):
                    array = numpy.array(array)
                if array.shape == ():
                    array = full(length, array)

                if length != array.shape[0]:
                    raise ValueError("array {0} has len {1} but other arrays have len {2}".format(repr(str(instruction.extern)), len(array), length))

                symbols[instruction.name] = array

            elif isinstance(instruction, histbook.instr.Assign):
                symbols[instruction.name] = histbook.calc.calculate(instruction.expr, symbols)

            elif isinstance(instruction, histbook.instr.Export):
                data = symbols[instruction.name]
                for i, j in instruction.destination:
                    self._destination[i][j] = data

            elif isinstance(instruction, histbook.instr.Delete):
                del symbols[instruction.name]

            else:
                raise AssertionError(instruction)

        return length



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math

import numpy

import histbook.axis
import histbook.instr
import histbook.expr
import histbook.hist

def isspark(arrays, more):
    out = arrays.__class__.__name__ == "DataFrame" and arrays.__class__.__module__ == "pyspark.sql.dataframe"
    if out and len(more) > 0:
        raise TypeError("if arrays is a PySpark DataFrame, keyword arguments are not allowed")
    return out

def tocolumns(df, expr):
    import pyspark.sql.functions as fcns

    if isinstance(expr, histbook.expr.Const):
        return fcns.lit(expr.value)

    elif isinstance(expr, (histbook.expr.Name, histbook.expr.Predicate)):
        return df[expr.value]

    elif isinstance(expr, histbook.expr.Call):
        if expr.fcn == "abs" or expr.fcn == "fabs":
            return fcns.abs(tocolumns(df, expr.args[0]))
        elif expr.fcn == "max" or expr.fcn == "fmax":
            return fcns.greatest(*[tocolumns(df, x) for x in expr.args])
        elif expr.fcn == "min" or expr.fcn == "fmin":
            return fcns.least(*[tocolumns(df, x) for x in expr.args])
        elif expr.fcn == "arccos":
            return fcns.acos(tocolumns(df, expr.args[0]))
        elif expr.fcn == "arccosh":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "arcsin":
            return fcns.asin(tocolumns(df, expr.args[0]))
        elif expr.fcn == "arcsinh":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "arctan2":
            return fcns.atan2(tocolumns(df, expr.args[0]), tocolumns(df, expr.args[1]))
        elif expr.fcn == "arctan":
            return fcns.atan(tocolumns(df, expr.args[0]))
        elif expr.fcn == "arctanh":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "ceil":
            return fcns.ceil(tocolumns(df, expr.args[0]))
        elif expr.fcn == "copysign":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "cos":
            return fcns.cos(tocolumns(df, expr.args[0]))
        elif expr.fcn == "cosh":
            return fcns.cosh(tocolumns(df, expr.args[0]))
        elif expr.fcn == "rad2deg":
            return tocolumns(df, expr.args[0]) * (180.0 / math.pi)
        elif expr.fcn == "erfc":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "erf":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "exp":
            return fcns.exp(tocolumns(df, expr.args[0]))
        elif expr.fcn == "expm1":
            return fcns.expm1(tocolumns(df, expr.args[0]))
        elif expr.fcn == "factorial":
            return fcns.factorial(tocolumns(df, expr.args[0]))
        elif expr.fcn == "floor":
            return fcns.floor(tocolumns(df, expr.args[0]))
        elif expr.fcn == "fmod":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "gamma":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "hypot":
            return fcns.hypot(tocolumns(df, expr.args[0]), tocolumns(df, expr.args[1]))
        elif expr.fcn == "isinf":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "isnan":
            return fcns.isnan(tocolumns(df, expr.args[0]))
        elif expr.fcn == "lgamma":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "log10":
            return fcns.log10(tocolumns(df, expr.args[0]))
        elif expr.fcn == "log1p":
            return fcns.log1p(tocolumns(df, expr.args[0]))
        elif expr.fcn == "log":
            return fcns.log(tocolumns(df, expr.args[0]))
        elif expr.fcn == "pow":
            return fcns.pow(tocolumns(df, expr.args[0]), tocolumns(df, expr.args[1]))
        elif expr.fcn == "deg2rad":
            return tocolumns(df, expr.args[0]) * (math.pi / 180.0)
        elif expr.fcn == "sinh":
            return fcns.sinh(tocolumns(df, expr.args[0]))
        elif expr.fcn == "sin":
            return fcns.sin(tocolumns(df, expr.args[0]))
        elif expr.fcn == "sqrt":
            return fcns.sqrt(tocolumns(df, expr.args[0]))
        elif expr.fcn == "tanh":
            return fcns.tanh(tocolumns(df, expr.args[0]))
        elif expr.fcn == "tan":
            return fcns.tan(tocolumns(df, expr.args[0]))
        elif expr.fcn == "trunc":
            raise NotImplementedError(expr.fcn)  # FIXME (fcns.trunc is for dates)
        elif expr.fcn == "xor":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "conjugate":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "exp2":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "heaviside":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "isfinite":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "left_shift" and isinstance(expr.args[1], histbook.expr.Const):
            return fcns.shiftLeft(tocolumns(df, expr.args[0]), expr.args[1].value)
        elif expr.fcn == "log2":
            return fcns.log2(tocolumns(df, expr.args[0]))
        elif expr.fcn == "logaddexp2":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "logaddexp":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "mod" or expr.fcn == "fmod":
            return tocolumns(df, expr.args[0]) % tocolumns(df, expr.args[1])
        elif expr.fcn == "right_shift" and isinstance(expr.args[1], histbook.expr.Const):
            return fcns.shiftRight(tocolumns(df, expr.args[0]), expr.args[1].value)
        elif expr.fcn == "rint":
            return fcns.rint(tocolumns(df, expr.args[0]))
        elif expr.fcn == "sign":
            raise NotImplementedError(expr.fcn)  # FIXME
        elif expr.fcn == "where":
            return fcns.when(tocolumns(df, expr.args[0]), tocolumns(df, expr.args[1])).otherwise(tocolumns(df, expr.args[2]))
        elif expr.fcn == "numpy.equal":
            return tocolumns(df, expr.args[0]) == tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.not_equal":
            return tocolumns(df, expr.args[0]) != tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.less":
            return tocolumns(df, expr.args[0]) < tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.less_equal":
            return tocolumns(df, expr.args[0]) <= tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.isin":
            return tocolumns(df, expr.args[0]) in tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.logical_not":
            return ~tocolumns(df, expr.args[0])
        elif expr.fcn == "numpy.add":
            return tocolumns(df, expr.args[0]) + tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.subtract":
            return tocolumns(df, expr.args[0]) - tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.multiply":
            return tocolumns(df, expr.args[0]) * tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.true_divide":
            return tocolumns(df, expr.args[0]) / tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.logical_or":
            return tocolumns(df, expr.args[0]) | tocolumns(df, expr.args[1])
        elif expr.fcn == "numpy.logical_and":
            return tocolumns(df, expr.args[0]) & tocolumns(df, expr.args[1])
        else:
            raise NotImplementedError(expr.fcn)

    else:
        raise AssertionError(expr)

def fillspark(hist, df):
    import pyspark.sql.functions as fcns

    indexes = []
    for axis in hist._group + hist._fixed:
        exprcol = tocolumns(df, histbook.instr.totree(axis._parsed))

        if isinstance(axis, histbook.axis.groupby):
            indexes.append(exprcol)

        elif isinstance(axis, histbook.axis.groupbin):
            scaled = (exprcol - float(axis.origin)) * (1.0/float(axis.binwidth))
            if axis.closedlow:
                discretized = fcns.floor(scaled)
            else:
                discretized = fcns.ceil(scaled) - 1
            indexes.append(fcns.nanvl(discretized * float(axis.binwidth) + float(axis.origin), fcns.lit("NaN")))

        elif isinstance(axis, histbook.axis.bin):
            scaled = (exprcol - float(axis.low)) * (int(axis.numbins) / (float(axis.high) - float(axis.low)))
            if axis.closedlow:
                discretized = fcns.floor(scaled) + 1
            else:
                discretized = fcns.ceil(scaled)
            indexes.append(fcns.when(fcns.isnull(exprcol) | fcns.isnan(exprcol), int(axis.numbins) + 2).otherwise(fcns.greatest(fcns.lit(0), fcns.least(fcns.lit(int(axis.numbins) + 1), discretized))))

        elif isinstance(axis, histbook.axis.intbin):
            indexes.append(fcns.greatest(fcns.lit(0), fcns.least(fcns.lit(int(axis.max) - int(axis.min) + 1), fcns.round(exprcol - int(axis.min) + 1))))

        elif isinstance(axis, histbook.axis.split):
            def build(x, i):
                if i < len(axis.edges):
                    if axis.closedlow:
                        return build(x.when(exprcol < float(axis.edges[i]), i), i + 1)
                    else:
                        return build(x.when(exprcol <= float(axis.edges[i]), i), i + 1)
                else:
                    return x.otherwise(i)
            indexes.append(build(fcns.when(fcns.isnull(exprcol) | fcns.isnan(exprcol), len(axis.edges) + 1), 0))

        elif isinstance(axis, histbook.axis.cut):
            indexes.append(fcns.when(exprcol, 0).otherwise(1))

        else:
            raise AssertionError(axis)

    aliasnum = [-1]
    def alias(x):
        aliasnum[0] += 1
        return x.alias("@" + str(aliasnum[0]))

    index = alias(fcns.struct(*indexes))

    selectcols = [index]
    if hist._weightoriginal is not None:
        weightcol = tocolumns(df, histbook.instr.totree(hist._weightparsed))
    for axis in hist._profile:
        exprcol = tocolumns(df, histbook.instr.totree(axis._parsed))
        if hist._weightoriginal is None:
            selectcols.append(alias(exprcol))
            selectcols.append(alias(exprcol*exprcol))
        else:
            selectcols.append(alias(exprcol*weightcol))
            selectcols.append(alias(exprcol*exprcol*weightcol))

    if hist._weightoriginal is None:
        df2 = df.select(*selectcols)
    else:
        selectcols.append(alias(weightcol))
        selectcols.append(alias(weightcol*weightcol))
        df2 = df.select(*selectcols)

    aggs = [fcns.sum(df2[n]) for n in df2.columns[1:]]
    if hist._weightoriginal is None:
        aggs.append(fcns.count(df2[df2.columns[0]]))

    def getornew(content, key, nextaxis):
        if key in content:
            return content[key]
        elif isinstance(nextaxis, histbook.axis.GroupAxis):
            return {}
        else:
            return numpy.zeros(hist._shape, dtype=histbook.hist.COUNTTYPE)

    def recurse(index, columns, axis, content):
        if len(axis) == 0:
            content += columns

        elif isinstance(axis[0], (histbook.axis.groupby, histbook.axis.groupbin)):
            content[index[0]] = recurse(index[1:], columns, axis[1:], getornew(content, index[0], axis[1] if len(axis) > 1 else None))
            if isinstance(axis[0], histbook.axis.groupbin) and None in content:
                content["NaN"] = content[None]
                del content[None]

        elif isinstance(axis[0], (histbook.axis.bin, histbook.axis.intbin, histbook.axis.split)):
            i = index[0] - (1 if not axis[0].underflow else 0)
            if int(i) < axis[0].totbins:
                recurse(index[1:], columns, axis[1:], content[int(i)])

        elif isinstance(axis[0], histbook.axis.cut):
            recurse(index[1:], columns, axis[1:], content[0 if index[0] else 1])

        else:
            raise AssertionError(axis[0])

        return content

    query = df2.groupBy(df2[df2.columns[0]]).agg(*aggs)

    def wait():
        for row in query.collect():
            recurse(row[0], row[1:], hist._group + hist._fixed, hist._content)

    return wait



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy

import histbook.axis

class Exportable(object):
    """Mix-in for methods that export histograms to third-party software."""

    def pandas(self, *axis, **opts):
        """
        Exports the data in the histogram to a Pandas DataFrame.

        Parameters
        ----------
        *axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            axis or axes to include in the table; if no axes or all :py:class:`profile <histbook.axis.profile>` axes, this function calls :py:meth:`Hist.table <histbook.proj.Projectable.table>`; if all :py:class:`cut <histbook.axis.cut>` axes, this function calls :py:meth:`Hist.fraction <histbook.proj.Projectable.fraction>`

        Keyword Arguments
        -----------------
        **opts : *any*
            passed to :py:meth:`Hist.table <histbook.proj.Projectable.table>` or :py:meth:`Hist.fraction <histbook.proj.Projectable.fraction>`; see these methods for options
        """
        import pandas as pd

        axis = [x if isinstance(x, histbook.axis.Axis) else self.axis[x] for x in axis]

        opts["recarray"] = True
        opts["columns"] = True

        if all(isinstance(x, histbook.axis.ProfileAxis) for x in axis):
            (content, columns) = self.table(*axis, **opts)
            allaxis = self._group + self._fixed

        elif all(isinstance(x, histbook.axis.cut) for x in axis):
            (content, columns), denomhist = self._fraction(axis, opts, True)
            allaxis = denomhist._group + denomhist._fixed

        else:
            raise TypeError("selected axis must be all profiles (for table) or all cuts (for fraction)")

        names = [None for x in allaxis]
        arrays = []
        keys = [[] for x in allaxis]

        def index(j, content, key):
            if j == len(allaxis):
                if len(self._fixed) == 0:
                    arrays.append(numpy.array([content]))

                for j, k in enumerate(key):
                    keys[j].append(k)

            else:
                axis = allaxis[j]
                names[j] = str(axis.expr)

                if isinstance(axis, histbook.axis.groupby):
                    for n in sorted(content):
                        index(j + 1, content[n], key + (n,))

                elif isinstance(axis, histbook.axis.groupbin):
                    closed = "left" if axis.closedlow else "right"
                    for n in sorted(content):
                        if n == "NaN":
                            index(j + 1, content[n], key + ("{NaN}",))
                        else:
                            index(j + 1, content[n], key + (pd.Interval(n, n + float(axis.binwidth), closed=closed),))

                else:
                    if content is not None:
                        arrays.append(content)

                    if isinstance(axis, histbook.axis.bin):
                        closed = "left" if axis.closedlow else "right"
                        if axis.underflow:
                            index(j + 1, None, key + (pd.Interval(float("-inf"), axis.low, closed=closed),))
                        last = axis.low
                        for i in range(axis.numbins):
                            this = (float(i + 1) / float(axis.numbins)) * float(axis.high - axis.low) + float(axis.low)
                            index(j + 1, None, key + (pd.Interval(last, this, closed=closed),))
                            last = this
                        if axis.overflow:
                            index(j + 1, None, key + (pd.Interval(axis.high, float("inf"), closed=closed),))
                        if axis.nanflow:
                            index(j + 1, None, key + ("{NaN}",))

                    elif isinstance(axis, histbook.axis.intbin):
                        if axis.underflow:
                            index(j + 1, None, key + (pd.Interval(float("-inf"), int(axis.min), closed="left"),))
                        for i in range(int(axis.min), int(axis.max) + 1):
                            index(j + 1, None, key + (str(i),))
                        if axis.overflow:
                            index(j + 1, None, key + (pd.Interval(int(axis.max), float("inf"), closed="right"),))

                    elif isinstance(axis, histbook.axis.split):
                        closed = "left" if axis.closedlow else "right"
                        if axis.underflow:
                            index(j + 1, None, key + (pd.Interval(float("-inf"), axis.edges[0], closed=closed),))
                        last = axis.edges[0]
                        for this in axis.edges[1:]:
                            index(j + 1, None, key + (pd.Interval(last, this, closed=closed),))
                            last = this
                        if axis.overflow:
                            index(j + 1, None, key + (pd.Interval(last, float("inf"), closed=closed),))
                        if axis.nanflow:
                            index(j + 1, None, key + ("{NaN}",))

                    elif isinstance(axis, histbook.axis.cut):
                        index(j + 1, None, key + (False,))
                        index(j + 1, None, key + (True,))

                    elif isinstance(axis, histbook.axis._nullaxis):
                        index(j + 1, None, key + ("",))

                    else:
                        raise AssertionError(axis)

        index(0, content, ())

        if len(arrays) == 0:
            return pd.DataFrame(index=pd.MultiIndex.from_arrays(keys, names=names),
                                columns=columns)
        else:
            arrays = numpy.concatenate(arrays)
            return pd.DataFrame(index=pd.MultiIndex.from_arrays(keys, names=names),
                                columns=arrays.dtype.names,
                                data=arrays.view(arrays.dtype[arrays.dtype.names[0]]).reshape(len(keys[0]), -1))

    def root(self, *axis, **opts):
        """
        Exports the data in the histogram to a ROOT histogram.

        The histogram may need to be selected (:py:meth:`Hist.select <histbook.proj.Projectable.select>`) or projected (:py:meth:`Hist.project <histbook.proj.Projectable.project>`) to make it useful in a ROOT histogram.

        Parameters
        ----------
        *axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            axis or axes to include in the output; if no axes, this function returns a histogram of counts; if one :py:class:`profile <histbook.axis.profile>`, this function returns a profile; ...

        Keyword Arguments
        -----------------
        name : string
            name to give to the ROOT object *(default is the empty string)*

        title : string
            title to give to the ROOT object *(default is the empty string)*

        cache : dict-like object
            if supplied, the return value is inserted into ``cache`` keyed by its name; this for convenience (ROOT objects must be kept in scope to be drawn)
        """
        import ROOT

        name = opts.pop("name", "")
        title = opts.pop("title", "")
        cache = opts.pop("cache", {})
        if len(opts) > 0:
            raise TypeError("unrecognized options for Hist.root: {0}".format(" ".join(opts)))

        if len(axis) == 0:
            axis = self._group + self._fixed
        axis = [x if isinstance(x, histbook.axis.Axis) else self.axis[x] for x in axis]
        for x in axis:
            if x not in self._group + self._fixed + self._profile:
                raise IndexError("no such axis: {0}".format(x))

        binaxis = []
        profile = None
        for x in axis:
            if isinstance(x, histbook.axis.ProfileAxis):
                if profile is None:
                    profile = x
                else:
                    raise ValueError("only one profile axis allowed: {0}, {1}".format(profile, x))
            else:
                binaxis.append(x)

        projected = self.project(*binaxis)
        if profile is None:
            content = projected.table(count=True, error=(self._weightparsed is not None))
        else:
            content = projected.table(profile, count=True, error=True)

        if len(binaxis) == 0:
            raise TypeError("cannot present zero-axis data in ROOT")

        elif len(binaxis) == 1 and isinstance(binaxis[0], histbook.axis.groupby):
            if profile is None:
                raise NotImplementedError("TH1 with string-labeled x-axis")

            else:
                raise NotImplementedError("TProfile with string-labeled x-axis")

        elif len(binaxis) == 2 and isinstance(binaxis[0], histbook.axis.groupby) and isinstance(binaxis[1], histbook.axis.groupby):
            if profile is None:
                raise NotImplementedError("TH2 with string-labeled x-axis and y-axis")

            else:
                raise NotImplementedError("TProfile2 with string-labeled x-axis and y-axis")

        elif len(binaxis) == 2 and isinstance(binaxis[0], histbook.axis.groupby):
            if profile is None:
                raise NotImplementedError("TH2 with string-labeled x-axis only")

            else:
                raise NotImplementedError("TProfile2 with string-labeled x-axis only")

        elif len(binaxis) == 2 and isinstance(binaxis[1], histbook.axis.groupby):
            if profile is None:
                raise NotImplementedError("TH2 with string-labeled y-axis only")

            else:
                raise NotImplementedError("TProfile2 with string-labeled y-axis only")

        elif any(isinstance(x, histbook.axis.groupby) for x in binaxis):
            raise TypeError("cannot present more than two categorical axes in ROOT")

        elif any(isinstance(x, histbook.axis.groupbin) for x in binaxis):
            if profile is None:
                raise NotImplementedError("THnSparse with fixed axes converted to sparse")

            else:
                raise TypeError("cannot present sparsely binned profile plots in ROOT")

        elif len(binaxis) == 1:
            xaxis = binaxis[0]

            if profile is None:
                out = ROOT.TH1D(name, title, xaxis.numbins, xaxis.low, xaxis.high)

                entries = numpy.zeros(xaxis.numbins + 2, dtype=numpy.float64)
                entries[(0 if xaxis.underflow else 1) : (None if xaxis.overflow else -1)] = content["count()"][: (-1 if xaxis.nanflow else None)]
                out.SetContent(entries)
                out.SetEntries(entries.sum())

                if "err(count())" in content.dtype.names:
                    errors = numpy.zeros(xaxis.numbins + 2, dtype=numpy.float64)
                    errors[(0 if xaxis.underflow else 1) : (None if xaxis.overflow else -1)] = content["err(count())"][: (-1 if xaxis.nanflow else None)]
                    out.SetError(errors)

            else:
                out = ROOT.TProfile(name, title, xaxis.numbins, xaxis.low, xaxis.high)

                entries = numpy.zeros(xaxis.numbins + 2, dtype=numpy.float64)
                entries[(0 if xaxis.underflow else 1) : (None if xaxis.overflow else -1)] = content[content.dtype.names[0]][: (-1 if xaxis.nanflow else None)]

                contents = numpy.zeros(xaxis.numbins + 2, dtype=numpy.float64)
                contents[(0 if xaxis.underflow else 1) : (None if xaxis.overflow else -1)] = content[content.dtype.names[2]][: (-1 if xaxis.nanflow else None)]

                errors = numpy.zeros(xaxis.numbins + 2, dtype=numpy.float64)
                errors[(0 if xaxis.underflow else 1) : (None if xaxis.overflow else -1)] = content[content.dtype.names[3]][: (-1 if xaxis.nanflow else None)]

                # you have to do crazy things to set TProfile bins by hand
                out.SetEntries(entries.sum())
                rootcontents = contents * entries
                rooterrors = numpy.sqrt(((errors**2 * entries) + contents**2) * entries)
                for i in range(xaxis.numbins + 2):
                    out.SetBinEntries(i, entries[i])
                    out.SetBinContent(i, rootcontents[i])
                    out.SetBinError(i, rooterrors[i])

        elif len(binaxis) == 2:
            if profile is None:
                raise NotImplementedError("TH2")

            else:
                raise NotImplementedError("TProfile2D")

        elif len(binaxis) == 3:
            if profile is None:
                raise NotImplementedError("TH3")

            else:
                raise NotImplementedError("TProfile3D")

        else:
            if profile is None:
                raise NotImplementedError("THn")

            else:
                raise TypeError("cannot present more than 3-dimensional profile plots in ROOT")

        cache[out.GetName()] = out
        return out



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pickle
import unittest

import numpy

from histbook.axis import *
from histbook.hist import *

class TestHist(unittest.TestCase):
    def runTest(self):
        pass

    def test_calc(self):
        h = Hist(bin("x + 0.1", 10, 0, 1))
        h.fill(x=[0.4, 0.3, 0.3, 0.5, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[0], [0], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [0]])

    def test_defs(self):
        h = Hist(bin("y", 10, 0, 1), defs={"y": "x + 0.1"})
        h.fill(x=[0.4, 0.3, 0.3, 0.5, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[0], [0], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [0]])

    def test_nofixed(self):
        h = Hist(groupby("c"))
        h.fill(c=["one", "two", "one", "one", "one"])
        self.assertEqual(h._content["one"].tolist(), [4])
        self.assertEqual(h._content["two"].tolist(), [1])

        h = Hist(groupby("c"), profile("x"))
        h.fill(c=["one", "two", "one", "one", "one"], x=[3, 2, 3, 5, 5])
        self.assertEqual(h._content["one"].tolist(), [16, 68, 4])
        self.assertEqual(h._content["two"].tolist(), [2, 4, 1])

        h = Hist(profile("x"))
        h.fill(x=[3, 3, 5, 5])
        self.assertEqual(h._content.tolist(), [16, 68, 4])

        h = Hist(weight="x")
        h.fill(x=[3, 3, 5, 5])
        self.assertEqual(h._content.tolist(), [16, 68])

        h = Hist(profile("y"), weight="x")
        h.fill(x=[2], y=[3])
        self.assertEqual(h._content.tolist(), [6, 18, 2, 4])

        h = Hist(bin("y", 1, -100, 100, underflow=False, overflow=False, nanflow=False), profile("y"), weight="x")
        h.fill(x=[2], y=[3])
        self.assertEqual(h._content.tolist(), [[6, 18, 2, 4]])

    def test_bin(self):
        h = Hist(bin("x", 10, 10, 11))
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8])
        self.assertEqual(h._content.tolist(), [[0], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [0], [0]])

        h = Hist(bin("x", 10, 0, 1))
        h.fill(x=[0.4, 0.3, 123, 99, 0.3, numpy.nan, numpy.nan, numpy.nan, 0.5, -99, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[1], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [2], [3]])

        h = Hist(bin("x", 10, 0, 1, underflow=False))
        h.fill(x=[0.4, 0.3, 123, 99, 0.3, numpy.nan, numpy.nan, numpy.nan, 0.5, -99, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [2], [3]])

        h = Hist(bin("x", 10, 0, 1, overflow=False))
        h.fill(x=[0.4, 0.3, 123, 99, 0.3, numpy.nan, numpy.nan, numpy.nan, 0.5, -99, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[1], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [3]])

        h = Hist(bin("x", 10, 0, 1, nanflow=False))
        h.fill(x=[0.4, 0.3, 123, 99, 0.3, numpy.nan, numpy.nan, numpy.nan, 0.5, -99, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[1], [0], [0], [0], [2], [2], [1], [0], [0], [1], [0], [2]])

        h = Hist(bin("x", 10, 0, 1, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[0.4, 0.3, 123, 99, 0.3, numpy.nan, numpy.nan, numpy.nan, 0.5, -99, 0.4, 0.8])
        self.assertEqual(h._content.tolist(), [[0], [0], [0], [2], [2], [1], [0], [0], [1], [0]])

        h = Hist(bin("x", 2, 0, 2))
        h.fill(x=[0.0, 0.0001, 0.0001, 0.5, 0.5, 0.5, 0.9999, 0.9999, 0.9999, 0.9999, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0001, 1.0001, 1.0001, 1.0001, 1.0001, 1.0001, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001])
        self.assertEqual(h._content.tolist(), [[0], [1 + 2 + 3 + 4], [5 + 6 + 7 + 8], [9 + 10], [0]])

        h = Hist(bin("x", 2, 0, 2, closedlow=False))
        h.fill(x=[0.0, 0.0001, 0.0001, 0.5, 0.5, 0.5, 0.9999, 0.9999, 0.9999, 0.9999, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0001, 1.0001, 1.0001, 1.0001, 1.0001, 1.0001, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 1.9999, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001, 2.0001])
        self.assertEqual(h._content.tolist(), [[1], [2 + 3 + 4 + 5], [6 + 7 + 8 + 9], [10], [0]])

    def test_bin_bin(self):
        h = Hist(bin("x", 3, 0, 3, underflow=False, overflow=False, nanflow=False), bin("y", 5, 0, 5, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[1], y=[3])
        self.assertEqual(h._content.tolist(), [[[0], [0], [0], [0], [0]], [[0], [0], [0], [1], [0]], [[0], [0], [0], [0], [0]]])

    def test_bin_weight(self):
        h = Hist(bin("x", 10, 10, 11), weight="y")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[0.1, 0.1, 0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.2, 0.020000000000000004], [0.2, 0.020000000000000004], [0.1, 0.010000000000000002], [0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

        h = Hist(bin("x", 10, 10, 11)).weight("y")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[0.1, 0.1, 0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.2, 0.020000000000000004], [0.2, 0.020000000000000004], [0.1, 0.010000000000000002], [0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

    def test_bin_filter(self):
        h = Hist(bin("x", 10, 10, 11), filter="y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-1, -1, -1, 1, 1, 1])
        self.assertEqual(h._content.tolist(), [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1], [1, 1], [0, 0], [0, 0], [1, 1], [0, 0], [0, 0], [0, 0]])

        h = Hist(bin("x", 10, 10, 11)).filter("y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-1, -1, -1, 1, 1, 1])
        self.assertEqual(h._content.tolist(), [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1], [1, 1], [0, 0], [0, 0], [1, 1], [0, 0], [0, 0], [0, 0]])

    def test_bin_weight_filter1(self):
        h = Hist(bin("x", 10, 10, 11), weight=2, filter="y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-1, -1, -1, 1, 1, 1])
        self.assertEqual(h._content.tolist(), [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [2, 4], [2, 4], [0, 0], [0, 0], [2, 4], [0, 0], [0, 0], [0, 0]])

        h = Hist(bin("x", 10, 10, 11)).weight(2).filter("y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-1, -1, -1, 1, 1, 1])
        self.assertEqual(h._content.tolist(), [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [2, 4], [2, 4], [0, 0], [0, 0], [2, 4], [0, 0], [0, 0], [0, 0]])

    def test_bin_weight_filter2(self):
        h = Hist(bin("x", 10, 10, 11), weight="y", filter="y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-0.1, -0.1, -0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.1, 0.010000000000000002], [0.1, 0.010000000000000002], [0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

        h = Hist(bin("x", 10, 10, 11)).weight("y").filter("y > 0")
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[-0.1, -0.1, -0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.1, 0.010000000000000002], [0.1, 0.010000000000000002], [0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

    def test_bin_big(self):
        numpy.random.seed(12345)
        xdata = numpy.round(numpy.random.normal(0, 1, 10000), 2)
        weights = numpy.random.uniform(-2, 2, 10000)
        both = numpy.random.randint(0, 9999, 10)
        xdata[numpy.random.randint(0, 9999, 10)] = numpy.nan
        weights[numpy.random.randint(0, 9999, 10)] = numpy.nan
        xdata[both] = numpy.nan
        weights[both] = numpy.nan

        nn = ~numpy.isnan(xdata)
        nnweights = weights.copy()
        nnweights[numpy.isnan(weights)] = 0.0
        nnweights2 = nnweights**2
        
        with numpy.errstate(invalid="ignore"):
            under = numpy.count_nonzero(xdata < 0.0)
            bin0 = numpy.count_nonzero(numpy.logical_and(0.0 <= xdata, xdata < 0.1))
            bin1 = numpy.count_nonzero(numpy.logical_and(0.1 <= xdata, xdata < 0.2))
            bin2 = numpy.count_nonzero(numpy.logical_and(0.2 <= xdata, xdata < 0.3))
            bin3 = numpy.count_nonzero(numpy.logical_and(0.3 <= xdata, xdata < 0.4))
            bin4 = numpy.count_nonzero(numpy.logical_and(0.4 <= xdata, xdata < 0.5))
            bin5 = numpy.count_nonzero(numpy.logical_and(0.5 <= xdata, xdata < 0.6))
            bin6 = numpy.count_nonzero(numpy.logical_and(0.6 <= xdata, xdata < 0.7))
            bin7 = numpy.count_nonzero(numpy.logical_and(0.7 <= xdata, xdata < 0.8))
            bin8 = numpy.count_nonzero(numpy.logical_and(0.8 <= xdata, xdata < 0.9))
            bin9 = numpy.count_nonzero(numpy.logical_and(0.9 <= xdata, xdata < 1.0))
            over = numpy.count_nonzero(xdata >= 1.0)
            nan = numpy.count_nonzero(numpy.isnan(xdata))

            wunder = numpy.sum(nnweights * numpy.logical_and(nn, xdata < 0.0))
            wbin0 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.0 <= xdata, xdata < 0.1)))
            wbin1 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.1 <= xdata, xdata < 0.2)))
            wbin2 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.2 <= xdata, xdata < 0.3)))
            wbin3 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.3 <= xdata, xdata < 0.4)))
            wbin4 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.4 <= xdata, xdata < 0.5)))
            wbin5 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.5 <= xdata, xdata < 0.6)))
            wbin6 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.6 <= xdata, xdata < 0.7)))
            wbin7 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.7 <= xdata, xdata < 0.8)))
            wbin8 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.8 <= xdata, xdata < 0.9)))
            wbin9 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.9 <= xdata, xdata < 1.0)))
            wover = numpy.sum(nnweights * numpy.logical_and(nn, (xdata >= 1.0)))
            wnan = numpy.sum(nnweights * numpy.isnan(xdata))

            w2under = numpy.sum(nnweights2 * numpy.logical_and(nn, (xdata < 0.0)))
            w2bin0 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.0 <= xdata, xdata < 0.1)))
            w2bin1 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.1 <= xdata, xdata < 0.2)))
            w2bin2 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.2 <= xdata, xdata < 0.3)))
            w2bin3 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.3 <= xdata, xdata < 0.4)))
            w2bin4 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.4 <= xdata, xdata < 0.5)))
            w2bin5 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.5 <= xdata, xdata < 0.6)))
            w2bin6 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.6 <= xdata, xdata < 0.7)))
            w2bin7 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.7 <= xdata, xdata < 0.8)))
            w2bin8 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.8 <= xdata, xdata < 0.9)))
            w2bin9 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.9 <= xdata, xdata < 1.0)))
            w2over = numpy.sum(nnweights2 * numpy.logical_and(nn, (xdata >= 1.0)))
            w2nan = numpy.sum(nnweights2 * numpy.isnan(xdata))

        for underflow in False, True:
            for overflow in False, True:
                for nanflow in False, True:
                    compare = [bin0, bin1, bin2, bin3, bin4, bin5, bin6, bin7, bin8, bin9]
                    if underflow:
                        compare.insert(0, under)
                    if overflow:
                        compare.append(over)
                    if nanflow:
                        compare.append(nan)
                    # h = Hist(bin("x", 10, 0, 1, underflow=underflow, overflow=overflow, nanflow=nanflow))
                    # h.fill(x=xdata)
                    # self.assertEqual(h._content.reshape(-1).tolist(), compare)

                    compare = [[wbin0, w2bin0], [wbin1, w2bin1], [wbin2, w2bin2], [wbin3, w2bin3], [wbin4, w2bin4], [wbin5, w2bin5], [wbin6, w2bin6], [wbin7, w2bin7], [wbin8, w2bin8], [wbin9, w2bin9]]
                    if underflow:
                        compare.insert(0, [wunder, w2under])
                    if overflow:
                        compare.append([wover, w2over])
                    if nanflow:
                        compare.append([wnan, w2nan])
                    h = Hist(bin("x", 10, 0, 1, underflow=underflow, overflow=overflow, nanflow=nanflow)).weight("w")
                    h.fill(x=xdata, w=weights)
                    self.assertTrue(numpy.absolute((h._content - numpy.array(compare)).reshape(-1)).max() < 1e-10)

        with numpy.errstate(invalid="ignore"):
            under = numpy.count_nonzero(xdata <= 0.0)
            bin0 = numpy.count_nonzero(numpy.logical_and(0.0 < xdata, xdata <= 0.1))
            bin1 = numpy.count_nonzero(numpy.logical_and(0.1 < xdata, xdata <= 0.2))
            bin2 = numpy.count_nonzero(numpy.logical_and(0.2 < xdata, xdata <= 0.3))
            bin3 = numpy.count_nonzero(numpy.logical_and(0.3 < xdata, xdata <= 0.4))
            bin4 = numpy.count_nonzero(numpy.logical_and(0.4 < xdata, xdata <= 0.5))
            bin5 = numpy.count_nonzero(numpy.logical_and(0.5 < xdata, xdata <= 0.6))
            bin6 = numpy.count_nonzero(numpy.logical_and(0.6 < xdata, xdata <= 0.7))
            bin7 = numpy.count_nonzero(numpy.logical_and(0.7 < xdata, xdata <= 0.8))
            bin8 = numpy.count_nonzero(numpy.logical_and(0.8 < xdata, xdata <= 0.9))
            bin9 = numpy.count_nonzero(numpy.logical_and(0.9 < xdata, xdata <= 1.0))
            over = numpy.count_nonzero(xdata > 1.0)
            nan = numpy.count_nonzero(numpy.isnan(xdata))

            wunder = numpy.sum(nnweights * numpy.logical_and(nn, xdata <= 0.0))
            wbin0 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.0 < xdata, xdata <= 0.1)))
            wbin1 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.1 < xdata, xdata <= 0.2)))
            wbin2 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.2 < xdata, xdata <= 0.3)))
            wbin3 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.3 < xdata, xdata <= 0.4)))
            wbin4 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.4 < xdata, xdata <= 0.5)))
            wbin5 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.5 < xdata, xdata <= 0.6)))
            wbin6 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.6 < xdata, xdata <= 0.7)))
            wbin7 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.7 < xdata, xdata <= 0.8)))
            wbin8 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.8 < xdata, xdata <= 0.9)))
            wbin9 = numpy.sum(nnweights * numpy.logical_and(nn, numpy.logical_and(0.9 < xdata, xdata <= 1.0)))
            wover = numpy.sum(nnweights * numpy.logical_and(nn, (xdata > 1.0)))
            wnan = numpy.sum(nnweights * numpy.isnan(xdata))

            w2under = numpy.sum(nnweights2 * numpy.logical_and(nn, (xdata <= 0.0)))
            w2bin0 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.0 < xdata, xdata <= 0.1)))
            w2bin1 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.1 < xdata, xdata <= 0.2)))
            w2bin2 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.2 < xdata, xdata <= 0.3)))
            w2bin3 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.3 < xdata, xdata <= 0.4)))
            w2bin4 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.4 < xdata, xdata <= 0.5)))
            w2bin5 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.5 < xdata, xdata <= 0.6)))
            w2bin6 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.6 < xdata, xdata <= 0.7)))
            w2bin7 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.7 < xdata, xdata <= 0.8)))
            w2bin8 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.8 < xdata, xdata <= 0.9)))
            w2bin9 = numpy.sum(nnweights2 * numpy.logical_and(nn, numpy.logical_and(0.9 < xdata, xdata <= 1.0)))
            w2over = numpy.sum(nnweights2 * numpy.logical_and(nn, (xdata > 1.0)))
            w2nan = numpy.sum(nnweights2 * numpy.isnan(xdata))

        for underflow in False, True:
            for overflow in False, True:
                for nanflow in False, True:
                    compare = [bin0, bin1, bin2, bin3, bin4, bin5, bin6, bin7, bin8, bin9]
                    if underflow:
                        compare.insert(0, under)
                    if overflow:
                        compare.append(over)
                    if nanflow:
                        compare.append(nan)
                    h = Hist(bin("x", 10, 0, 1, underflow=underflow, overflow=overflow, nanflow=nanflow, closedlow=False))
                    h.fill(x=xdata)
                    self.assertEqual(h._content.reshape(-1).tolist(), compare)

                    compare = [[wbin0, w2bin0], [wbin1, w2bin1], [wbin2, w2bin2], [wbin3, w2bin3], [wbin4, w2bin4], [wbin5, w2bin5], [wbin6, w2bin6], [wbin7, w2bin7], [wbin8, w2bin8], [wbin9, w2bin9]]
                    if underflow:
                        compare.insert(0, [wunder, w2under])
                    if overflow:
                        compare.append([wover, w2over])
                    if nanflow:
                        compare.append([wnan, w2nan])
                    h = Hist(bin("x", 10, 0, 1, underflow=underflow, overflow=overflow, nanflow=nanflow, closedlow=False)).weight("w")
                    h.fill(x=xdata, w=weights)
                    self.assertTrue(numpy.absolute((h._content - numpy.array(compare)).reshape(-1)).max() < 1e-10)

    def test_intbin(self):
        h = Hist(intbin("x", 0, 10, underflow=True, overflow=True))
        h.fill(x=range(-5, 15))
        self.assertEqual(h._content.tolist(), [[5], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [4]])

        h = Hist(intbin("x", 0, 10, underflow=True, overflow=False))
        h.fill(x=range(-5, 15))
        self.assertEqual(h._content.tolist(), [[5], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1]])

        h = Hist(intbin("x", 0, 10, underflow=False, overflow=True))
        h.fill(x=range(-5, 15))
        self.assertEqual(h._content.tolist(), [[1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [4]])

        h = Hist(intbin("x", 0, 10, underflow=False, overflow=False))
        h.fill(x=range(-5, 15))
        self.assertEqual(h._content.tolist(), [[1], [1], [1], [1], [1], [1], [1], [1], [1], [1], [1]])

    def test_split(self):
        h = Hist(split("x", 3))
        h.fill(x=[0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(h._content.tolist(), [[3], [4], [0]])

        h = Hist(split("x", 3))
        h.fill(x=[0, 1, 2, 3, 4, 5, numpy.nan])
        self.assertEqual(h._content.tolist(), [[3], [3], [1]])

        h = Hist(split("x", (2.5, 3.5)))
        h.fill(x=[0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(h._content.tolist(), [[3], [1], [3], [0]])

        h = Hist(split("x", (2.5, 3.5, 5.0)))
        h.fill(x=[0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(h._content.tolist(), [[3], [1], [1], [2], [0]])

        h = Hist(split("x", (2.5, 3.5, 5.0), closedlow=False))
        h.fill(x=[0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(h._content.tolist(), [[3], [1], [2], [1], [0]])

        under = [[2]]
        over = [[4]]
        nan = [[1]]
        for underflow in (False, True):
            for overflow in (False, True):
                for nanflow in (False, True):
                    if underflow or overflow or nanflow:
                        h = Hist(split("x", 3, underflow=underflow, overflow=overflow, nanflow=nanflow))
                        h.fill(x=[numpy.nan, 1, 2, 3, 4, 5, 6])
                        self.assertEqual(h._content.tolist(), (under if underflow else []) + (over if overflow else []) + (nan if nanflow else []))

        under = [[2]]
        over = [[1]]
        nan = [[1]]
        for underflow in (False, True):
            for overflow in (False, True):
                for nanflow in (False, True):
                    h = Hist(split("x", (2.5, 3.5, 5.0), underflow=underflow, overflow=overflow, nanflow=nanflow, closedlow=False))
                    h.fill(x=[numpy.nan, 1, 2, 3, 4, 5, 6])
                    self.assertEqual(h._content.tolist(), (under if underflow else []) + [[1], [2]] + (over if overflow else []) + (nan if nanflow else []))

    def test_cut(self):
        h = Hist(cut("p"))
        h.fill(p=[False, True, True])
        self.assertEqual(h._content.tolist(), [[1], [2]])

        h = Hist(cut("x > 3"))
        h.fill(x=[1, 2, 3, 4, 5])
        self.assertEqual(h._content.tolist(), [[3], [2]])

        h = Hist(cut("x > 3 and p"))
        h.fill(x=[1, 2, 3, 4, 5], p=[True, False, True, False, True])
        self.assertEqual(h._content.tolist(), [[4], [1]])

        h = Hist(cut("x > 3 and y % 2 == 0"))
        h.fill(x=[1, 2, 3, 4, 5], y=[0, 1, 2, 3, 4])
        self.assertEqual(h._content.tolist(), [[4], [1]])

        h = Hist(cut("x > 3 and y % 2 == 0"), split("x", 3.5))
        h.fill(x=[1, 2, 3, 4, 5], y=[0, 1, 2, 3, 4])
        self.assertEqual(h._content.tolist(), [[[3], [1], [0]], [[0], [1], [0]]])

    def test_profile(self):
        h = Hist(bin("x", 10, 10, 11), profile("y"))
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[0.1, 0.1, 0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.2, 0.020000000000000004, 2.0], [0.2, 0.020000000000000004, 2.0], [0.1, 0.010000000000000002, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])

        h = Hist(bin("x", 10, 10, 11), profile("y"), profile("2*y"))
        h.fill(x=[10.4, 10.3, 10.3, 10.5, 10.4, 10.8], y=[0.1, 0.1, 0.1, 0.1, 0.1, 1.0])
        self.assertEqual(h._content.tolist(), [[0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.2, 0.020000000000000004, 0.4, 0.08000000000000002, 2.0], [0.2, 0.020000000000000004, 0.4, 0.08000000000000002, 2.0], [0.1, 0.010000000000000002, 0.2, 0.04000000000000001, 1.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [1.0, 1.0, 2.0, 4.0, 1.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0]])

    def test_groupby(self):
        h = Hist(groupby("c"), bin("x", 3, 1.0, 4.0, underflow=False, overflow=False, nanflow=False))
        h.fill(c=["one", "two", "three", "two", "one", "one", "one"], x=[1, 2, 3, 2, 1, 1, 3])
        self.assertEqual(h._content["one"].tolist(), [[3], [0], [1]])
        self.assertEqual(h._content["two"].tolist(), [[0], [2], [0]])
        self.assertEqual(h._content["three"].tolist(), [[0], [0], [1]])

    def test_groupby_groupby(self):
        h = Hist(groupby("c1"), groupby("c2"), bin("x", 1, 1.0, 2.0, underflow=False, overflow=False, nanflow=False))
        h.fill(c1=["one", "two", "one", "two"], c2=["uno", "uno", "dos", "dos"], x=[1, 1, 1, 1])
        self.assertEqual(h._content["one"]["uno"].tolist(), [[1]])
        self.assertEqual(h._content["two"]["uno"].tolist(), [[1]])
        self.assertEqual(h._content["one"]["dos"].tolist(), [[1]])
        self.assertEqual(h._content["two"]["dos"].tolist(), [[1]])

    def test_groupbin(self):
        h = Hist(groupbin("x", 10.0), bin("y", 4, 1.0, 5.0, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[0, 10, 15, 20], y=[1, 2, 3, 4])
        self.assertEqual(h._content[0.0].tolist(),  [[1], [0], [0], [0]])
        self.assertEqual(h._content[10.0].tolist(), [[0], [1], [1], [0]])
        self.assertEqual(h._content[20.0].tolist(), [[0], [0], [0], [1]])

        h = Hist(groupbin("x", 10.0, origin=1.0), bin("y", 4, 1.0, 5.0, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[0, 10, 15, 20], y=[1, 2, 3, 4])
        self.assertEqual(h._content[-9.0].tolist(), [[1], [0], [0], [0]])
        self.assertEqual(h._content[1.0].tolist(),  [[0], [1], [0], [0]])
        self.assertEqual(h._content[11.0].tolist(), [[0], [0], [1], [1]])

    def test_groupby_groupbin(self):
        h = Hist(bin("y", 4, 1.0, 5.0, underflow=False, overflow=False, nanflow=False), groupby("c"), groupbin("x", 10.0))
        h.fill(c=["one", "one", "two", "two"], x=[0, 10, 15, 20], y=[1, 2, 3, 4])
        self.assertEqual(h._content["one"][0.0].tolist(),  [[1], [0], [0], [0]])
        self.assertEqual(h._content["one"][10.0].tolist(), [[0], [1], [0], [0]])
        self.assertEqual(h._content["two"][10.0].tolist(), [[0], [0], [1], [0]])
        self.assertEqual(h._content["two"][20.0].tolist(), [[0], [0], [0], [1]])

        h = Hist(groupbin("x", 10.0), bin("y", 4, 1.0, 5.0, underflow=False, overflow=False, nanflow=False), groupby("c"))
        h.fill(c=["one", "one", "two", "two"], x=[0, 10, 15, 20], y=[1, 2, 3, 4])
        self.assertEqual(h._content[0.0]["one"].tolist(),  [[1], [0], [0], [0]])
        self.assertEqual(h._content[10.0]["one"].tolist(), [[0], [1], [0], [0]])
        self.assertEqual(h._content[10.0]["two"].tolist(), [[0], [0], [1], [0]])
        self.assertEqual(h._content[20.0]["two"].tolist(), [[0], [0], [0], [1]])

    def test_pickle(self):
        h = Hist(split("x", (1, 2, 3)), bin("y", 10, 0, 1), defs={"y": "x + 0.1"}, weight="sqrt(x)", filter="x > 2")
        self.assertEqual(h, pickle.loads(pickle.dumps(h)))

        h.fill(x=[1, 2, 3])
        self.assertEqual(h, pickle.loads(pickle.dumps(h)))

    def test_json(self):
        h = Hist(split("x", (1, 2, 3)), bin("y", 10, 0, 1), defs={"y": "x + 0.1"}, weight="sqrt(x)", filter="x > 2")
        self.assertEqual(h, Hist.fromjson(h.tojson()))

        h.fill(x=[1, 2, 3])
        self.assertEqual(h, Hist.fromjson(h.tojson()))

